@@ -14,19 +14,26 @@ from boto3 import Session
 from botocore.session import Session as botocoreSession
 from botocore.auth import SigV4Auth
 from botocore.awsrequest import AWSRequest
-from gremlin_python.driver import client
+from gremlin_python.driver import client, serializer
 from gremlin_python.driver.protocol import GremlinServerError
-from neo4j import GraphDatabase
+from neo4j import GraphDatabase, DEFAULT_DATABASE
+from neo4j.exceptions import AuthError
+from base64 import b64encode
 import nest_asyncio
-# from graph_notebook.magics.graph_magic import NEPTUNE_CONFIG_HOST_IDENTIFIERS
+
 
 # This patch is no longer needed when graph_notebook is using the a Gremlin Python
 # client >= 3.5.0 as the HashableDict is now part of that client driver.
 # import graph_notebook.neptune.gremlin.graphsonV3d0_MapType_objectify_patch  # noqa F401
 
+DEFAULT_GREMLIN_SERIALIZER = 'graphsonv3'
+DEFAULT_GREMLIN_TRAVERSAL_SOURCE = 'g'
 DEFAULT_SPARQL_CONTENT_TYPE = 'application/x-www-form-urlencoded'
 DEFAULT_PORT = 8182
 DEFAULT_REGION = 'us-east-1'
+DEFAULT_NEO4J_USERNAME = 'neo4j'
+DEFAULT_NEO4J_PASSWORD = 'password'
+DEFAULT_NEO4J_DATABASE = DEFAULT_DATABASE
 
 NEPTUNE_SERVICE_NAME = 'neptune-db'
 logger = logging.getLogger('client')
@@ -92,26 +99,59 @@ STREAM_PG = 'PropertyGraph'
 STREAM_RDF = 'RDF'
 STREAM_ENDPOINTS = {STREAM_PG: 'gremlin', STREAM_RDF: 'sparql'}
 
-NEPTUNE_CONFIG_HOST_IDENTIFIERS = ["amazonaws.com"]
+NEPTUNE_CONFIG_HOST_IDENTIFIERS = ["neptune.amazonaws.com", "neptune.*.amazonaws.com.cn"]
+
+false_str_variants = [False, 'False', 'false', 'FALSE']
+
+GRAPHSONV3_VARIANTS = ['graphsonv3', 'graphsonv3d0', 'graphsonserializersv3d0']
+GRAPHSONV2_VARIANTS = ['graphsonv2', 'graphsonv2d0', 'graphsonserializersv2d0']
+GRAPHBINARYV1_VARIANTS = ['graphbinaryv1', 'graphbinary', 'graphbinaryserializersv1']
+
+STATISTICS_MODES = ["status", "disableAutoCompute", "enableAutoCompute", "refresh", "delete"]
+STATISTICS_LANGUAGE_INPUTS = ["propertygraph", "pg", "gremlin", "sparql", "rdf"]
 
 
 def is_allowed_neptune_host(hostname: str, host_allowlist: list):
     for host_snippet in host_allowlist:
-        if host_snippet in hostname:
+        if re.search(host_snippet, hostname):
             return True
     return False
 
 
+def get_gremlin_serializer(serializer_str: str):
+    serializer_lower = serializer_str.lower()
+    if serializer_lower == 'graphbinaryv1':
+        return serializer.GraphBinarySerializersV1()
+    elif serializer_lower == 'graphsonv2':
+        return serializer.GraphSONSerializersV2d0()
+    else:
+        return serializer.GraphSONSerializersV3d0()
+
+
 class Client(object):
-    def __init__(self, host: str, port: int = DEFAULT_PORT, ssl: bool = True, region: str = DEFAULT_REGION,
-                 sparql_path: str = '/sparql', gremlin_traversal_source: str = 'g', auth=None, session: Session = None,
+    def __init__(self, host: str, port: int = DEFAULT_PORT, ssl: bool = True, ssl_verify: bool = True,
+                 region: str = DEFAULT_REGION, sparql_path: str = '/sparql',
+                 gremlin_traversal_source: str = DEFAULT_GREMLIN_TRAVERSAL_SOURCE,
+                 gremlin_username: str = '', gremlin_password: str = '',
+                 gremlin_serializer: str = DEFAULT_GREMLIN_SERIALIZER,
+                 neo4j_username: str = DEFAULT_NEO4J_USERNAME, neo4j_password: str = DEFAULT_NEO4J_PASSWORD,
+                 neo4j_auth: bool = True, neo4j_database: str = DEFAULT_NEO4J_DATABASE,
+                 auth=None, session: Session = None,
                  proxy_host: str = '', proxy_port: int = DEFAULT_PORT,
                  neptune_hosts: list = None):
         self.target_host = host
         self.target_port = port
         self.ssl = ssl
+        self.ssl_verify = ssl_verify
         self.sparql_path = sparql_path
         self.gremlin_traversal_source = gremlin_traversal_source
+        self.gremlin_username = gremlin_username
+        self.gremlin_password = gremlin_password
+        self.gremlin_serializer = get_gremlin_serializer(gremlin_serializer)
+        self.neo4j_username = neo4j_username
+        self.neo4j_password = neo4j_password
+        self.neo4j_auth = neo4j_auth
+        self.neo4j_database = neo4j_database
         self.region = region
         self._auth = auth
         self._session = session
@@ -183,7 +223,7 @@ class Client(object):
 
         uri = f'{self._http_protocol}://{self.host}:{self.port}{sparql_path}'
         req = self._prepare_request('POST', uri, data=data, headers=headers)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def sparql(self, query: str, headers=None, explain: str = '', path: str = '') -> requests.Response:
@@ -223,7 +263,9 @@ class Client(object):
         ws_url = f'{self.get_uri_with_port(use_websocket=True)}/gremlin'
         request = self._prepare_request('GET', ws_url)
         traversal_source = 'g' if self.is_neptune_domain() else self.gremlin_traversal_source
-        return client.Client(ws_url, traversal_source, headers=dict(request.headers), **transport_kwargs)
+        return client.Client(ws_url, traversal_source, username=self.gremlin_username,
+                             password=self.gremlin_password, message_serializer=self.gremlin_serializer,
+                             headers=dict(request.headers), **transport_kwargs)
 
     def gremlin_query(self, query, transport_args=None, bindings=None):
         if transport_args is None:
@@ -252,7 +294,7 @@ class Client(object):
         uri = f'{self.get_uri_with_port()}/gremlin'
         data = {'gremlin': query}
         req = self._prepare_request('POST', uri, data=json.dumps(data), headers=headers)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def gremlin_status(self, query_id: str = '', include_waiting: bool = False):
@@ -279,33 +321,57 @@ class Client(object):
             for param, value in args.items():
                 data[param] = value
         req = self._prepare_request('POST', url, data=json.dumps(data))
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def opencypher_http(self, query: str, headers: dict = None, explain: str = None) -> requests.Response:
         if headers is None:
             headers = {}
 
-        if 'content-type' not in headers:
-            headers['content-type'] = 'application/x-www-form-urlencoded'
+        url = f'{self._http_protocol}://{self.host}:{self.port}/'
 
-        url = f'{self._http_protocol}://{self.host}:{self.port}/openCypher'
-        data = {
-            'query': query
-        }
-        if explain:
-            data['explain'] = explain
-            headers['Accept'] = "text/html"
+        if self.is_neptune_domain():
+            if 'content-type' not in headers:
+                headers['content-type'] = 'application/x-www-form-urlencoded'
+            url += 'openCypher'
+            data = {
+                'query': query
+            }
+            if explain:
+                data['explain'] = explain
+                headers['Accept'] = "text/html"
+        else:
+            url += 'db/neo4j/tx/commit'
+            headers['content-type'] = 'application/json'
+            headers['Accept'] = 'application/vnd.neo4j.jolt+json-seq'
+
+            data_dict = {
+                "statements": [
+                    {
+                        "statement": query
+                    }
+                ]
+            }
+            data = json.dumps(data_dict)
+            if self.neo4j_auth:
+                user_and_pass = self.neo4j_username + ":" + self.neo4j_password
+                user_and_pass_base64 = b64encode(user_and_pass.encode())
+                headers['authorization'] = user_and_pass_base64
 
         req = self._prepare_request('POST', url, data=data, headers=headers)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def opencyper_bolt(self, query: str, **kwargs):
         driver = self.get_opencypher_driver()
-        with driver.session() as session:
-            res = session.run(query, kwargs)
-            data = res.data()
+        with driver.session(database=self.neo4j_database) as session:
+            try:
+                res = session.run(query, kwargs)
+                data = res.data()
+            except AuthError:
+                print("Neo4J Bolt request failed with an authentication error. Please ensure that the 'neo4j' section "
+                      "of your %graph_notebook_config contains the correct credentials and auth setting.")
+                data = []
         driver.close()
         return data
 
@@ -321,23 +387,33 @@ class Client(object):
 
         return self._query_status('openCypher', query_id=query_id, cancelQuery=True, silent=silent)
 
-    def get_opencypher_driver(self, user: str = 'neo4j', password: str = 'password'):
+    def get_opencypher_driver(self):
         url = f'bolt://{self.host}:{self.port}'
 
-        if self._session:
-            method = 'POST'
-            headers = {
-                'HttpMethod': method,
-                'Host': url
-            }
-            aws_request = self._get_aws_request('POST', url)
-            for item in aws_request.headers.items():
-                headers[item[0]] = item[1]
+        if self.is_neptune_domain():
+            user = DEFAULT_NEO4J_USERNAME
+            if self._session:
+                method = 'POST'
+                headers = {
+                    'HttpMethod': method,
+                    'Host': url
+                }
+                aws_request = self._get_aws_request('POST', url)
+                for item in aws_request.headers.items():
+                    headers[item[0]] = item[1]
 
-            auth_str = json.dumps(headers)
-            password = auth_str
+                auth_str = json.dumps(headers)
+                password = auth_str
+            else:
+                password = DEFAULT_NEO4J_PASSWORD
+            auth_final = (user, password)
+        else:
+            if self.neo4j_auth:
+                auth_final = (self.neo4j_username, self.neo4j_password)
+            else:
+                auth_final = None
 
-        driver = GraphDatabase.driver(url, auth=(user, password), encrypted=self.ssl)
+        driver = GraphDatabase.driver(url, auth=auth_final, encrypted=self.ssl)
         return driver
 
     def stream(self, url, **kwargs) -> requests.Response:
@@ -345,13 +421,13 @@ class Client(object):
         for k, v in kwargs.items():
             params[k] = v
         req = self._prepare_request('GET', url, params=params,data='')
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res.json()
 
     def status(self) -> requests.Response:
         url = f'{self._http_protocol}://{self.host}:{self.port}/status'
         req = self._prepare_request('GET', url, data='')
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def load(self, source: str, source_format: str, iam_role_arn: str = None, **kwargs) -> requests.Response:
@@ -375,7 +451,7 @@ class Client(object):
         url = f'{self._http_protocol}://{self.host}:{self.port}/loader'
         raw = json.dumps(payload)
         req = self._prepare_request('POST', url, data=raw, headers={'content-type': 'application/json'})
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def load_status(self, load_id: str = '', **kwargs) -> requests.Response:
@@ -388,14 +464,14 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/loader'
         req = self._prepare_request('GET', url, params=params)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def cancel_load(self, load_id: str) -> requests.Response:
         url = f'{self._http_protocol}://{self.host}:{self.port}/loader'
         params = {'loadId': load_id}
         req = self._prepare_request('DELETE', url, params=params)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def initiate_reset(self) -> requests.Response:
@@ -404,7 +480,7 @@ class Client(object):
         }
         url = f'{self._http_protocol}://{self.host}:{self.port}/system'
         req = self._prepare_request('POST', url, data=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def perform_reset(self, token: str) -> requests.Response:
@@ -414,7 +490,7 @@ class Client(object):
         }
         url = f'{self._http_protocol}://{self.host}:{self.port}/system'
         req = self._prepare_request('POST', url, data=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def dataprocessing_start(self, s3_input_uri: str, s3_output_uri: str, **kwargs) -> requests.Response:
@@ -428,7 +504,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/dataprocessing'
         req = self._prepare_request('POST', url, data=json.dumps(data), headers={'content-type': 'application/json'})
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def dataprocessing_job_status(self, job_id: str, neptune_iam_role_arn: str = '') -> requests.Response:
@@ -437,7 +513,7 @@ class Client(object):
         if neptune_iam_role_arn != '':
             data['neptuneIamRoleArn'] = neptune_iam_role_arn
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def dataprocessing_list(self, max_items: int = 10, neptune_iam_role_arn: str = '') -> requests.Response:
@@ -449,7 +525,7 @@ class Client(object):
         if neptune_iam_role_arn != '':
             data['neptuneIamRoleArn'] = neptune_iam_role_arn
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def dataprocessing_stop(self, job_id: str, clean=False, neptune_iam_role_arn: str = '') -> requests.Response:
@@ -461,7 +537,7 @@ class Client(object):
             data['neptuneIamRoleArn'] = neptune_iam_role_arn
 
         req = self._prepare_request('DELETE', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltraining_start(self, data_processing_job_id: str, train_model_s3_location: str,
@@ -483,7 +559,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltraining'
         req = self._prepare_request('POST', url, data=json.dumps(data), headers={'content-type': 'application/json'})
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltraining_list(self, max_items: int = 10, neptune_iam_role_arn: str = '') -> requests.Response:
@@ -496,14 +572,14 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltraining'
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltraining_job_status(self, training_job_id: str, neptune_iam_role_arn: str = '') -> requests.Response:
         data = {} if neptune_iam_role_arn == '' else {'neptuneIamRoleArn': neptune_iam_role_arn}
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltraining/{training_job_id}'
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltraining_stop(self, training_job_id: str, neptune_iam_role_arn: str = '',
@@ -517,7 +593,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltraining/{training_job_id}'
         req = self._prepare_request('DELETE', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltransform_create(self, output_s3_location: str, dataprocessing_job_id: str = '',
@@ -551,7 +627,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltransform'
         req = self._prepare_request('POST', url, data=json.dumps(data), headers=headers)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltransform_status(self, job_id: str, iam_role: str = '') -> requests.Response:
@@ -561,7 +637,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltransform/{job_id}'
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltransform_list(self, iam_role: str = '', max_items: int = 10) -> requests.Response:
@@ -574,7 +650,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltransform'
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def modeltransform_stop(self, job_id: str, iam_role: str = '', clean: bool = False) -> requests.Response:
@@ -586,7 +662,7 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/modeltransform/{job_id}'
         req = self._prepare_request('DELETE', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def endpoints_create(self, model_training_job_id: str = '', model_transform_job_id: str = '',
@@ -606,21 +682,21 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/endpoints'
         req = self._prepare_request('POST', url, data=json.dumps(data), headers={'content-type': 'application/json'})
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def endpoints_status(self, endpoint_id: str, neptune_iam_role_arn: str = '') -> requests.Response:
         data = {} if neptune_iam_role_arn == '' else {'neptuneIamRoleArn': neptune_iam_role_arn}
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/endpoints/{endpoint_id}'
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def endpoints_delete(self, endpoint_id: str, neptune_iam_role_arn: str = '') -> requests.Response:
         data = {} if neptune_iam_role_arn == '' else {'neptuneIamRoleArn': neptune_iam_role_arn}
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/endpoints/{endpoint_id}'
         req = self._prepare_request('DELETE', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def endpoints(self, max_items: int = 10, neptune_iam_role_arn: str = '') -> requests.Response:
@@ -632,21 +708,21 @@ class Client(object):
 
         url = f'{self._http_protocol}://{self.host}:{self.port}/ml/endpoints'
         req = self._prepare_request('GET', url, params=data)
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def export(self, host: str, params: dict, ssl: bool = True) -> requests.Response:
         protocol = 'https' if ssl else 'http'
         url = f'{protocol}://{host}/{EXPORT_ACTION}'
         req = self._prepare_request('POST', url, data=json.dumps(params), service="execute-api")
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def export_status(self, host, job_id, ssl: bool = True) -> requests.Response:
         protocol = 'https' if ssl else 'http'
         url = f'{protocol}://{host}/{EXPORT_ACTION}/{job_id}'
         req = self._prepare_request('GET', url, service="execute-api")
-        res = self._http_session.send(req)
+        res = self._http_session.send(req, verify=self.ssl_verify)
         return res
 
     def _query_status(self, language: str, *, query_id: str = '', **kwargs) -> requests.Response:
@@ -661,7 +737,27 @@ class Client(object):
             'Content-Type': 'application/x-www-form-urlencoded'
         }
         url = f'{self._http_protocol}://{self.host}:{self.port}/{language}/status'
+        requests.get()
         req = self._prepare_request('POST', url, data=data, headers=headers)
+        res = self._http_session.send(req, verify=self.ssl_verify)
+        return res
+
+    def statistics(self, language: str, mode: str = '') -> requests.Response:
+        headers = {
+            'Accept': 'application/json'
+        }
+        if language in ["pg", "gremlin"]:
+            language = "propertygraph"
+        elif language == "rdf":
+            language = "sparql"
+        url = f'{self._http_protocol}://{self.host}:{self.port}/{language}/statistics'
+        if mode in ['', 'status']:
+            req = self._prepare_request('GET', url, headers=headers)
+        elif mode == 'delete':
+            req = self._prepare_request('DELETE', url, headers=headers)
+        else:
+            data = {'mode': mode}
+            req = self._prepare_request('POST', url, data=json.dumps(data), headers=headers)
         res = self._http_session.send(req)
         return res
 
@@ -675,6 +771,7 @@ class Client(object):
             aws_request = self._get_aws_request(method=method, url=url, data=data, params=params, headers=headers,
                                                 service=service)
             request.headers = dict(aws_request.headers)
+
         return request.prepare()
 
     def _get_aws_request(self, method, url, *, data=None, params=None, headers=None, service=NEPTUNE_SERVICE_NAME):
@@ -734,11 +831,31 @@ class ClientBuilder(object):
     def with_gremlin_traversal_source(self, traversal_source: str):
         self.args['gremlin_traversal_source'] = traversal_source
         return ClientBuilder(self.args)
+        
+    def with_gremlin_login(self, username: str, password: str):
+        self.args['gremlin_username'] = username
+        self.args['gremlin_password'] = password
+        return ClientBuilder(self.args)
+
+    def with_gremlin_serializer(self, message_serializer: str):
+        self.args['gremlin_serializer'] = message_serializer
+        return ClientBuilder(self.args)
+    
+    def with_neo4j_login(self, username: str, password: str, auth: bool, database: str):
+        self.args['neo4j_username'] = username
+        self.args['neo4j_password'] = password
+        self.args['neo4j_auth'] = auth
+        self.args['neo4j_database'] = database
+        return ClientBuilder(self.args)
 
     def with_tls(self, tls: bool):
         self.args['ssl'] = tls
         return ClientBuilder(self.args)
 
+    def with_ssl_verify(self, ssl_verify: bool):
+        self.args['ssl_verify'] = ssl_verify
+        return ClientBuilder(self.args)
+
     def with_region(self, region: str):
         self.args['region'] = region
         return ClientBuilder(self.args)
